
=head1 Subroutines to create a tcl object
=head2 Description
These subroutines provide tools for compiling a tcl source file into
a corresponding object file that can be later parsed by the tcl linker.

  Copyright (c) 1994 by C*A*T*T.  All Rights Reserved.
  File:   tclstrip
  Author: Richard Schwaninger
  Date:   30/08/1994
  System: ithron.catt.co.at
  RCS:    $Id: compile.pl,v 1.2 1996/02/16 06:53:29 risc Exp $


=cut

# library to compile tcl files

%Require=();
%Locate=();


=head1 FilesToList - read in all files
=head2 Synopsis
	FilesToList(\@lst,@fnames)
=head2 Description
This routine reads in all the files given as arguments and splits them
into a big array of lines so we can process them as one junk.

=head2 Arguments
=over 3
=item lst
Name of a list variable where the read lines are stored
=item fnames
This is a list of file names.
=back
=head2 Return Value
If all the files are readable then this routine returns the name of
the variable that contains the lines, otherwise it dies with an
appropriate error-message.
=cut

sub FilesToList {
   local($lst,@fnames)=@_;
   local($t)="";
   foreach $i (@fnames) {
      if ( ! -r $i) {
	 die "Compile: '$i' not readable\n";
      }
      &Verbose("reading file '$i'");
      # read file in a big junk;
      open(IN,"$i");
      local($sep)=$/;
      undef $/;
      $t .=<IN>;
      $/=$sep;
      close(IN);
   }

   @$lst=split(/\n/,$t);
   return $lst;
}

=head1 Defines - handle defines
=head2 Synopsis
	Defines(@src,\@dst)
=head2 Description
this piece of code handles C<#defines> (much like the C praeprocessor
does). It creates a list of defined (or undefined values) for the
C<CheckCondition> function defined later.

=head2 Arguments
=over 3
=item src
This is a list of lines from the original source file
=item dst
This is a list of lines for returning the (possibly modified) lines
to the caller. 

=back
=head2 Return Value
The result is returned in the list C<dst> passed as argument, the
return value itself is undefined.

=cut

sub Defines {
   local($src,$dst)=@_;
   &Verbose("processing defines");
   foreach $line (@$src) {
      # print "D: $line\n";
      if($line =~ /^\s*#\s*define\s+(\S+)\s+(.*)/) {
	 $Defines{$1}=$2;
	 next;
      }
      if($line =~ /^\s*#\s*define\s+(\S+)/) {
	 $Defines{$1}=1;
	 next;
      }
      if($line =~ /^\s*#\s*undef\s+(\S+)/) {
	 if(defined($Defines{$1})) {
	    delete($Defines{$1});
	 }
	 next;
      }
      push(@$dst,$line);
   }
}

=head1 CheckCondition - conditional processing
=head2 Synopsis
	CheckCondition(@inp,\@out)
=head2 Description
This subroutine returns the source file (acutally a list of source
lines) but modifies the input according to conditions present in
the source (C<#ifdefs> like in the C preprocessor).

=cut

sub CheckCondition {
   local($line,$Lst,$LstElse)=@_;
   local($o);
   # print "CC: $line\n";
   if($line =~ /^\s*\#\s*else/) {
      # print "else: '$$Lst[$#$Lst]', $#$Lst\n";
      push(@$LstElse,$$Lst[$#$Lst]);
      next;
   }
   if($line =~ /^\s*\#\s*endif/) {
      $o=pop(@$Lst);
      # print "endif: ((($o)))'$$Lst[$#$Lst]', $#$Lst\n";
      # print "htelse:$$LstElse[$#$LstElse-1]\n";
      if($o eq $$LstElse[$#$LstElse]) {
	 # print "pop else too, $$LstElse[$#$LstElse]\n";
	 pop(@$LstElse);
      }
      next;
   }
   if($line =~ /^\s*#\s*ifdef\s+(\S+)/) {
      push(@$Lst,$1);
      # print "ifdef '$1': '$$Lst[$#$Lst]', $#$Lst $Defines{$1}\n";
      next;
   }
}


# remove conditional code
sub Conditionals {
   &Verbose("processing conditionals");
   local($src,$dst)=@_;
   local(@Lst)=();
   local(@LstElse)=();
   local($val,$ifd,$els);
   # NOTE: only one define at a time...
   foreach $line (@$src) {
      # print "C: $line\n";
      if($#Lst>=0) {
	 $ifd=$Lst[$#Lst];
	 # print "ifd==$ifd, ";
	 $val=$Defines{$ifd};
	 # print "val==$val\n";
	 if($#LstElse>=0) {
	    $els=$LstElse[$LstElse];
	 } else {
	    $els="";
	 }
      } else {
	 $ifd="";
      }
      &CheckCondition($line,\@Lst,\@LstElse);
      if ($ifd ne "") {
	 if($ifd eq $els) {
	    if(!$val) {
	       push(@$dst,$line);
	    }
	    next;
	 } else {
	    if($val) {
	       push(@$dst,$line);
	    }
	    next;
	 }
      }
      push(@$dst,$line);
   }
}


# remove all comments
sub Comments {
   &Verbose("removing comments");
   local($src,$dst)=@_;
   foreach $line (@$src) {
      #print "#: $line\n";
      if($line =~ /^\#\s+RCS:\s+\$Id/) {
	 # leave rcs line alone
	 push(@$dst,$line);
	 next;
      }
      if($line =~ /^\s*#/) {
	 next;
      }
      push(@$dst,$line);
   }
}

=head1 Indent - remove indentation
=head2 Synopsis
	Indent($src,$dst)
=head2 Description
This function removes any indentation that is not relevant to the
code. This speeds up processing a bit and makes source files smaller.
=head2 Arguments
=over 3
=item $src
This is a pointer to the actual list of lines for the input
=item $dst
This is a pointer to where the output lines should be put.
=back
=head2 Return Value
This function puts it's result into the output list and returns no
usefull value.

=cut

# remove all indentation
sub Indent {
   &Verbose("removing indentation");
   local($src,$dst)=@_;
   foreach $line (@$src) {
      if($line =~ /^\s+(.*)/) {
	 push(@$dst,$1);
      } else {
	 push(@$dst,$line);
      }
   }
}


# concat split lines

sub SplitLines {
   &Verbose("removing split lines");
   local($src,$dst)=@_;
   local($tmp)="";
   foreach $line (@$src) {
      if($line =~ /\\$/) {
	 chop($line);
	 if($line =~ /\S$/) {
	    $tmp .= "$line ";
	 } else {
	    $tmp .= $line;
	 }
	 next;
      } else {
	 $tmp .= $line;
	 push(@$dst,$tmp);
	 $tmp="";
      }
   }
}


# remove all indentation
sub EmptyLines {
   &Verbose("removing empty lines");
   local($src,$dst)=@_;
   foreach $line (@$src) {
      if($line !~ /^$/) {
	 push(@$dst,$line);
      }
   }
}


sub OpenParen {
   &Verbose("processing open braces");
   local($src,$dst)=@_;
   local($tmp)="";
   foreach $line (@$src) {
      if($line =~ /^(.*\{)\s*$/) {
	 $tmp .= $1;
      } else {
	 $tmp .= $line;
	 push(@$dst,$tmp);
	 $tmp="";
      }
   }
}


sub CloseParen {
   &Verbose("processing close braces");
   local($src,$dst)=@_;
   local($tmp)="";
   foreach $line (@$src) {
      if($line =~ /^\s*(\}.*)$/) {
	 $$dst[$#$dst] .= $1;
      } else {
	 push(@$dst,$line);
      }
   }
}


sub NoteRequire {
   &Verbose("noting requirements");
   local($src)=@_;
   local($lib,$version);
   local($class)=0;
   foreach $line (@$src) {
      if($class) {
	 if($line=~ /^\s*\}/) {
	    $class=0;
	    next;
	 }
	 if($line=~ /\s*(\S+)\s+([0-9\.]+)/) {
	    $lib=$1;
	    $version=$2;
	    $Require{$lib}=$version;
	 }
	 next;
      }
      # print "R: $line\n";
      if($line =~ /^\s*Require \{/) {
	 # a class require
	 $class=1;
	 next;
      }
      if($line =~ /Require\s+([A-Za-z_]+)\s+([0-9\.]+)/) {
	 $lib=$1;
	 $version=$2;
      } elsif($line =~ /Require\s+([A-Za-z_]+)/) {
	 warn "Require without version ('$line')\n";
	 $version=1.0;
      } elsif($line =~ /Require/) {
	 warn "stray Require in line '$line'\n";
	 next;
      } else {
	 next;
      }
      $Require{$lib}=$version;
   }
}


=head1 Compile - main compile routine
=head2 Description
This is the only routine that should be called from outside of this
module. It takes a list of files, processes them and writes them 
to the already open channel named C<OUT>.
=head2 Arguments
=over 3
=item $Outfile
This is the name of the output file (a stream to this file is
assumed to be open under the name C<OUT>).
=item @InFiles
This is a list of input file names. The files will be read in turn
and then processed.
=back
=head2 Return Value
does not return anything usefull
=cut

sub Compile {
   local($Outfile,@InFiles)=@_;

   FilesToList(\@src0,@InFiles);

   if($Optimize) {
      &Defines(\@src0,\@src1);
      undef @src0;
      &Conditionals(\@src1,\@src2);
      undef @src1;
      &Comments(\@src2,\@src3);
      undef @src2;
      &Indent(\@src3,\@src4);
      undef @src3;

      &NoteRequire(\@src4);

      &SplitLines(\@src4,\@src5);
      undef @src4;
      &EmptyLines(\@src5,\@src6);
      undef @src5;
      &OpenParen(\@src6,\@src7);
      undef @src6;
      &CloseParen(\@src7,\@srcX);
      undef @src7;
   } else {
      &Defines(\@src0,\@src1);
      undef @src0;
      &Conditionals(\@src1,\@src2);
      undef @src1;
      &Comments(\@src2,\@srcX);
      undef @src2;

      &NoteRequire(\@srcX);

   }

   $ftime=&filetime($InFile);
   $fname=&filetail($InFile);
   $path=&filedirname($InFile);
   $compiler=$longversion;
   print OUT <<EOT
# +OBJ
# COMPILER=$compiler
# VERSION=$version
# FILE=$fname
# PATH=$path
# MTIME=$ftime
# USER=$User
# DATE=$Date
# OPTIMIZE=$Optimize
# DEBUG=$Debug
EOT
    ;
   foreach $i (keys %Require) {
      print OUT "# RQ $i $Require{$i}\n";
   }
   print OUT "# +END\n";

   foreach $line (@srcX) {
      print OUT "$line\n";
   }
   print OUT "# +EOF\n";
   close(IN);
}


# EOF




















































